import logging
from io import BytesIO
from random import randint

from tcpchan.core.chan import Channel
from tcpchan.core.evt import (ChannelClosed, ChannelCreated, DataTransmit,
                              HandshakeFailed, HandshakeSuccess)
from tcpchan.core.msg import (ChannelPayload, CloseChannelRequest,
                              CreateChannelRequest, HandshakeReply,
                              HandshakeRequest, TCPChanMessage)

CONN_STATE_IDLE = 0
CONN_STATE_SETUP = 1
CONN_STATE_ESTABLISHED = 2
CONN_STATE_HANDSHAKE = 3
CONN_STATE_HANDSHAKE_SUCCESS = 4
CONN_STATE_HANDSHAKE_FAIL = 5

HANDSHAKE_MAGIC = 0xFEEDBACC


class BaseConnection:
    """ Base class for connection
    """

    def __init__(self, logger=None, event_callback=None):
        self._channels = {}
        self._buf = BytesIO()
        self._events = []
        self._state = CONN_STATE_IDLE
        self._event_callback = event_callback

        if self._event_callback and not callable(self._event_callback):
            raise ValueError("Expect callable event_callback.")

        if logger is not None:
            self._logger = logger
        else:
            self._logger = logging.getLogger(self.__class__.__name__)

    def connection_established(self):
        """ Called when underlying connection is established
        """
        raise NotImplementedError

    def connection_closed(self, exc):
        """ Called when underlying connection is closed

            Arguements:
                exc (Exception): exception indicating what happend
        """
        raise NotImplementedError

    def create_channel(self, channel_id=None):
        raise NotImplementedError

    def close_channel(self, channel_id=0):
        raise NotImplementedError

    def data_received(self, data):
        """ Called when underlying connection recevied new packet
            
            Arguments:
                data (bytes, memoryview): new data
        """
        raise NotImplementedError

    def next_event(self):
        """ Get next event from event queue
        """
        try:
            return self._events.pop(0)
        except IndexError:
            return None

    def add_events(self, events=[]):
        """ Add a list of events to event queue
        """
        self._events += events
        self.event_notify()

    def event_notify(self):
        """ Called when event is enqueued
        """
        if self._event_callback:
            self._event_callback()

    def close(self):
        raise NotImplementedError


class Connection(BaseConnection):
    """ TCPChan connection

        Attributes:
            channel_factory (callable): Factory function for channel creation.
            handshake_magic (int): Magic number to use during handshake
    """

    def __init__(self, channel_factory, handshake_magic=None, *arg, **kwargs):
        super().__init__(*arg, **kwargs)

        self._handlers = {
            CreateChannelRequest: self._handle_create_channel_request,
            CloseChannelRequest: self._handle_close_channel_request,
            ChannelPayload: self._handle_channel_payload,
            HandshakeRequest: self._handle_handshake_request,
            HandshakeReply: self._handle_handshake_reply,
        }

        if handshake_magic is None:
            self._magic = HANDSHAKE_MAGIC
        else:
            self._magic = handshake_magic

        self._channel_factory = channel_factory
        if not callable(self._channel_factory):
            raise ValueError("channel_factory must be callable.")

    def connection_established(self):
        """ Called on connection establishment
        """
        self._logger.debug("connection established.")
        self._state = CONN_STATE_ESTABLISHED

    def data_received(self, data):
        """ Called on data reception from network 
        """
        self._buf.write(data)

        offset = 0
        while True:
            try:
                msg, processed = TCPChanMessage.from_bytes(
                    self._buf.getbuffer()[offset:]
                )
            except ValueError:
                break

            try:
                self._handlers[msg.__class__](msg)
            except KeyError:
                self._logger.error('Unhandled message type "%s".', type(msg).__name__)
                self.close()

            offset += processed

        if offset > 0:
            newbuf = BytesIO()
            newbuf.write(self._buf.getbuffer()[offset:])
            self._buf = newbuf

    def channel_transmit_data(self, channel_id, data):
        payload = ChannelPayload(Channel=channel_id, Payload=data)

        self.add_events([DataTransmit(payload=payload.pack())])
        self._logger.debug("Scheduled data transmission from channel %d.", channel_id)

    def get_channel(self, channel_id):
        return self._channels.get(channel_id, None)

    def create_channel(self, channel_id=None):
        if not channel_id:
            channel_id = randint(0, 2 ** 32)

        while channel_id in self._channels:
            channel_id += 1

        msg = CreateChannelRequest(Channel=channel_id)
        self.add_events([DataTransmit(payload=msg.pack())])

        return self._create_channel(channel_id)

    def _create_channel(self, channel_id):
        if channel_id in self._channels:
            raise Exception(f"Duplicated channel id {channel_id}.")

        new_channel = self._channel_factory()
        new_channel.set_channel_id(channel_id)
        new_channel.set_connection(self)
        new_channel.channel_created()
        self._channels[channel_id] = new_channel

        self.add_events([ChannelCreated(channel_id=channel_id, channel=new_channel)])

        return new_channel

    def close_channel(self, channel_id):
        if self._delete_channel(channel_id):
            msg = CloseChannelRequest(Channel=channel_id)
            self.add_events(
                [
                    DataTransmit(payload=msg.pack()),
                    ChannelClosed(channel_id=channel_id),
                ]
            )

    def _delete_channel(self, channel_id):
        try:
            channel = self._channels[channel_id]
            del self._channels[channel_id]

            channel.close()
            return True
        except KeyError:
            self._logger.error("Deleting a non-existed channel %d.", channel_id)
            return False

    def _handle_create_channel_request(self, msg):
        self._logger.debug("handling create channel request.")
        self._create_channel(msg.Channel)

    def _handle_close_channel_request(self, msg):
        self._logger.debug("handling close channel request.")
        self._delete_channel(msg.Channel)

    def _handle_channel_payload(self, msg):
        self._logger.debug("handling channel payload.")
        try:
            self._channels[msg.Channel].data_received(msg.Payload)
        except KeyError:
            self._logger.error("Non-existed channel %d.", msg.Channel)

    def _handle_handshake_request(self, msg):
        self._logger.debug("handling handshake request.")

        if msg.Magic != self._magic:
            self._logger.debug("handshake failed due to mismatched magic.")
            self._state = CONN_STATE_HANDSHAKE_FAIL
            self.add_events([HandshakeFailed(reason="Mismatched magic.")])
            return

        self._logger.debug("handshake succeeded, sending reply.")
        msg = HandshakeReply(Magic=self._magic)
        self.add_events(
            [DataTransmit(payload=msg.pack()), HandshakeSuccess(),]
        )
        self._state = CONN_STATE_HANDSHAKE_SUCCESS

    def _handle_handshake_reply(self, msg):
        self._logger.debug("handling handshake reply.")

        if msg.Magic == self._magic:
            self._logger.debug("hanshake succeeded.")
            self._state = CONN_STATE_HANDSHAKE_SUCCESS
            evt = HandshakeSuccess()
        else:
            self._logger.debug("handshake failed due to mismatched magic.")
            self._state = CONN_STATE_HANDSHAKE_FAIL
            evt = HandshakeFailed(reason="Mismatched magic.")

        self.add_events([evt])


class ClientConnection(Connection):
    def connection_established(self):
        super().connection_established()
        msg = HandshakeRequest(Magic=self._magic)
        self.add_events([DataTransmit(payload=msg.pack())])
        self._state = CONN_STATE_HANDSHAKE


class ServerConnection(Connection):
    def connection_established(self):
        super().connection_established()
        self._state = CONN_STATE_HANDSHAKE


__all__ = ["Connection", "ClientConnection", "ServerConnection"]



from fpack import Bytes, Message, Uint8, Uint32, field_factory

TCPCHAN_OP_HANDSHAKE_REQUEST = 1
TCPCHAN_OP_HANDSHAKE_REPLY = 2
TCPCHAN_OP_CREATE_CHANNEL_REQUEST = 3
TCPCHAN_OP_CLOSE_CHANNEL_REQUEST = 4
TCPCHAN_OP_CHANNEL_PAYLOAD = 5


class BaseTCPChanMessage(Message):
    """ Base class for TCPChan messages

        Attributes:
            Fields (list): list of field
            version (int): protocol version (0)
    """

    Fields = [
        field_factory("Version", Uint8),
        field_factory("Op", Uint8),
    ]
    version = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.Op = self.opcode
        self.Version = self.version


class TCPChanMessage(BaseTCPChanMessage):
    @classmethod
    def from_bytes(self, data):
        msg, processed = super().from_bytes(data)

        if msg is None:
            return None, 0

        if msg.Op == TCPCHAN_OP_HANDSHAKE_REQUEST:
            return HandshakeRequest.from_bytes(data)

        if msg.Op == TCPCHAN_OP_HANDSHAKE_REPLY:
            return HandshakeReply.from_bytes(data)

        if msg.Op == TCPCHAN_OP_CREATE_CHANNEL_REQUEST:
            return CreateChannelRequest.from_bytes(data)

        if msg.Op == TCPCHAN_OP_CLOSE_CHANNEL_REQUEST:
            return CloseChannelRequest.from_bytes(data)

        if msg.Op == TCPCHAN_OP_CHANNEL_PAYLOAD:
            return ChannelPayload.from_bytes(data)


class HandshakeRequest(BaseTCPChanMessage):
    """ Handshake Request Message

        Handshake request message is sent as the first message to start the handshake procedure.
    """

    opcode = TCPCHAN_OP_HANDSHAKE_REQUEST
    Fields = TCPChanMessage.Fields + [
        field_factory("Magic", Uint32),
    ]


class HandshakeReply(BaseTCPChanMessage):
    """ Handshake Reply Message

        Handshake reply message is sent upon reception of handshake request.
    """

    opcode = TCPCHAN_OP_HANDSHAKE_REPLY
    Fields = TCPChanMessage.Fields + [
        field_factory("Magic", Uint32),
    ]


class ChannelMessage(Message):
    Fields = TCPChanMessage.Fields + [
        field_factory("Channel", Uint32),
    ]


class CreateChannelRequest(BaseTCPChanMessage):
    """ Create Channel Request Message

        Create channel request message is sent to request channel creation at the other end of the connection
    """

    opcode = TCPCHAN_OP_CREATE_CHANNEL_REQUEST
    Fields = ChannelMessage.Fields


class CloseChannelRequest(BaseTCPChanMessage):
    """ Close Channel Request Message

        Close channel request is sent to notify the other end of the connection a closed channel.
    """

    opcode = TCPCHAN_OP_CLOSE_CHANNEL_REQUEST
    Fields = ChannelMessage.Fields


class ChannelPayload(BaseTCPChanMessage):
    """ Channel Payload Message

        Channel payload message is used for delivery of payload to a specific channel.
    """

    opcode = TCPCHAN_OP_CHANNEL_PAYLOAD
    Fields = ChannelMessage.Fields + [
        field_factory("Payload", Bytes),
    ]


__all__ = [
    "TCPChanMessage",
    "HandshakeRequest",
    "HandshakeReply",
    "CreateChannelRequest",
    "CloseChannelRequest",
    "ChannelPayload",
]



import logging


class Channel:
    """ TCPChan Channel

        The channel instance is responsible for maintaining the logical
        channels in TCPChan connection.

        Attributes:
            connection (TCPChan.core.Connection): associated TCPChan connection
            channel_id (int): id of the channel
            logger (logging.Logger): logging utility
    """

    def __init__(self, connection=None, channel_id=0, logger=None):
        self._channel_id = channel_id
        self._conn = connection
        self._closed = False

        if logger:
            self._logger = logger
        else:
            self._logger = logging.getLogger(self.__class__.__name__)

    @property
    def channel_id(self):
        return self._channel_id

    @property
    def connection(self):
        return self._conn

    def set_connection(self, connection):
        """ Associate the channel with given connection
        """
        self._conn = connection

    def set_channel_id(self, channel_id):
        """ Set the id of the channel
        """
        self._channel_id = channel_id

    def channel_created(self):
        """ Called when channel is created
        """
        self._logger.debug("Channel %d is created.", self._channel_id)

    def write_data(self, data):
        """ Send channel data to connection

            Arguments:
                data (bytes): data to send to the other end of channel
        """
        if self._closed:
            self._logger.warn("Writing data to a closed channel.")
            return

        self._conn.channel_transmit_data(self._channel_id, data)

    def data_received(self, data):
        """ Called on reception of data
            
            Arguments:
                data (bytes): received data
        """
        raise NotImplementedError

    @property
    def is_closed(self):
        """ Tell whether channel is close or not
        """
        return self._closed

    def close(self):
        """ Close the channel
        """
        if not self._closed:
            self._closed = True

            if self._conn.get_channel(self.channel_id):
                self._conn.close_channel(self.channel_id)


__all__ = ["Channel"]



#!/usr/bin/env python

import asyncio
import logging

from tcpchan.core.conn import ClientConnection, ServerConnection
from tcpchan.core.evt import (ChannelClosed, ChannelCreated, DataTransmit,
                              HandshakeFailed, HandshakeSuccess)


class TCPChanBaseProtocol(asyncio.Protocol):
    """ TCP Channel Base Protocol

        Base protocol for the TCPChannelServerProtocol and TCPChannelClientProtocol
    
        Attributes:
            channel_factory (callable): a factory function to create new channel.
            logger (logging.Logger): optional, logging utility
            handshake_magic (int): optional, magic number to use during handshake
    """

    def __init__(
        self, channel_factory, logger=None, handshake_magic=None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)

        if logger:
            self._logger = logger
        else:
            self._logger = logging.getLogger(type(self).__name__)

        if not callable(channel_factory):
            raise TypeError("channel factory must be a callable.")

        self._handshake_magic = handshake_magic
        self._channel_factory = channel_factory

    def _event_handler(self):
        while True:
            ev = self._tcpchan.next_event()
            self._logger.debug("Event %s received.", ev)

            if ev is None:
                break

            if type(ev) == DataTransmit:
                self._transport.write(ev.payload)

            elif type(ev) == ChannelCreated:
                self.channel_created(ev.channel)

            elif type(ev) == HandshakeSuccess:
                self.handshake_success()

            elif type(ev) == ChannelClosed:
                self.channel_closed(ev.channel_id)

            elif type(ev) == HandshakeFailed:
                self.handshake_failed(self, reason=ev.reason)

    def connection_made(self, transport):
        self._transport = transport
        self._tcpchan.connection_established()

    def data_received(self, data):
        data = memoryview(data)
        self._tcpchan.data_received(data)

    def create_channel(self):
        """ Create new channel
        
            Calling this method creates a logical transmssion channel, and
            send a channel creation request to the other end of the connection.

            Returns:
                new_channel (Channel): Newly created channel
        """
        new_channel = self._tcpchan.create_channel()
        self._logger.debug("Channel %d is created.", new_channel.channel_id)

        return new_channel

    def handshake_success(self):
        """ Called when handshake success
        """

    def handshake_failed(self, reason):
        """ Called when handshake failed

            Arguments:
                reason (str): reason of handshake failure
        """

    def channel_created(self, channel):
        """ Called when channel is created from the other end of the connection

            Arguments:
                channel (Channel): newly created channel.
        """

    def channel_closed(self, channel_id):
        """ Called when channel is closed from the other end of the connection.

            Arguments:
                channel_id (int): the id of the closed channel
        """


class TCPChanServerProtocol(TCPChanBaseProtocol):
    """ TCPChan Server Protocol
        
        TCPChan protocol for server side.

        Attributes:
            channel_factory (callable): a factory function to create new channel.
            logger (logging.Logger): optional, logging utility
            handshake_magic (int): optional, magic number to use during handshake
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tcpchan = ServerConnection(
            self._channel_factory,
            event_callback=self._event_handler,
            handshake_magic=self._handshake_magic,
        )


class TCPChanClientProtocol(TCPChanBaseProtocol):
    """ TCPChan Client Protocol

        TCPChan protocol for client side.

        Attributes:
            channel_factory (callable): a factory function to create new channel.
            logger (logging.Logger): optional, logging utility
            handshake_magic (int): optional, magic number to use during handshake
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tcpchan = ClientConnection(
            self._channel_factory,
            event_callback=self._event_handler,
            handshake_magic=self._handshake_magic,
        )


__all__ = ["TCPChanClientProtocol", "TCPChanServerProtocol"]



#!/usr/bin/env python

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = "0.1.0"

setuptools.setup(
    name="tcpchan",
    version=version,
    author="Frank Chang",
    author_email="frank@csie.io",
    description="tcpchan is a TCP (de)multiplexer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/frankurcrazy/tcpchan",
    packages=setuptools.find_packages(),
    license="BSD",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: System :: Networking",
    ],
    keywords=["tcp", "multiplexer", "mux", "stream", "channel"],
    python_requires=">=3.7",
    install_requires=["fpack>=1.0.0",],
)



import typing
from dataclasses import dataclass


@dataclass
class BaseEvent:
    """ Base class for TCPChan events
    """


@dataclass
class HandshakeSuccess(BaseEvent):
    """ Handshake Success Event

        On handshake success event, client and server connection
        can start creating channel and tranmitting application data.
    """


@dataclass
class HandshakeFailed(BaseEvent):
    """ Handshake Failed Event

        When hanshake failed event is received, the connection should be
        shutdown immediately, the failure reason is indicated in ```reason``` field.
    """

    reason: str


@dataclass
class DataTransmit(BaseEvent):
    """ Data Transmit Event

        Data transmit event indicate the need for TX operation. The payload to be
        transmitted is indicated in the ```payload``` field.
    """

    payload: bytes


@dataclass
class ChannelClosed(BaseEvent):
    """ Channel Closed Event
        
        The event is received when a specific channel is closed. The id of the closed
        channel is indicated in the ```channel_id``` field.
    """

    channel_id: int


@dataclass
class ChannelCreated(BaseEvent):
    """ Channel Created Event
        
        The event is received upon successful creation of a channel. The id of the created
        channel and the channel instance itself are indicated in the ```channel_id``` and ```channel``` field
        respectively..
    """

    channel_id: int
    channel: typing.Any
